/*
 * Rmessage library. By Dave Lowell.
 *
 * This library puts a very thin layer on top of UDP sockets routines.
 * This layer augments traditional UDP by providing reliable, ordered,
 * delivery of datagram messages. All acknowledgements are done 
 * asynchronously, and an attempt is made to piggyback acks on return 
 * messages as appropriate. 
 */

/**
 ** Includes
 **/

#include <sys/time.h>
#include <signal.h>
#include "rmessage.h"
#include "hash.h"


/**
 ** Global Variables
 **/

hash_table	*connections;
hash_table	*seqno_table;
hash_table	*addr_table;
outstanding_msg	*outs_h;
outstanding_msg	*outs_t;
connection_info	*to_ack[RM_MAX_DESCRIPTORS];
int		curr_ack_index;
int		descriptors[RM_MAX_DESCRIPTORS+1];
int		msgs_on_sock[RM_MAX_DESCRIPTORS];
int		local_ports[RM_MAX_DESCRIPTORS];
int		num_desc = 0;


/**
 ** Support Functions
 **/

void* alloc(int size)
{
	void	*p;

	p = malloc(size);
	if (p == NULL) {
		fprintf(stderr, "rmessage: out of memory.\n");
		exit(1);
	}
	else return p;
}

void rm_lock(void)
{
	int	res;
	sigset_t m;

	sigemptyset(&m);
	sigaddset(&m, RM_TIMER_SIG);
	res = sigprocmask(SIG_BLOCK, &m, NULL);
	if (res < 0)
		perror("rm_lock");
}

void rm_unlock(void)
{
	int	res;
	sigset_t m;

	sigemptyset(&m);
	sigaddset(&m, RM_TIMER_SIG);
	res = sigprocmask(SIG_UNBLOCK, &m, NULL);
	if (res < 0)
		perror("rm_lock");
}

/*
 * get_address() converts its host argument string (either a hostname or an
 * IP address in dotted decimal format), into an internet address and returns
 * it, or FALSE on failure.
 */     
unsigned long get_address(char* host)
{
	struct in_addr  *addr_ptr; 
	unsigned long   inaddr;

	/*              
	 * First check if host string we've been passed is in
	 * dotted decimal format. Failing that, we'll get the 
	 * IP from the system.
	 */     
	if ((int)(inaddr = inet_addr(host)) == INADDR_NONE) {
		struct hostent  *hp;
		struct in_addr	*addr;

		hp = gethostbyname(host);
		if (hp == NULL) {
			return FALSE;
		}
		addr = (struct in_addr*) hp->h_addr;
		return addr->s_addr;
	}
	else {
		return inaddr;
	}
}


int get_local_port(int sock)
{
	int	port, res, len;
	struct sockaddr_in name;

	port = local_ports[sock];
	if (port == 0) {
		len = sizeof(name);
		res = getsockname(sock, (struct sockaddr*) &name, &len);
		if (res < 0) {
			perror("getsockname");
			return -1;
		}
		port = ntohs(name.sin_port);
		if (port == 0) {
			/*
			 * This socket has not yet been bound to a port.
			 * We'll just go ahead and bind it to a system
			 * assigned port, and then request the port num from
			 * the system.
			 */
			res = rm_bind(sock, 0);
			if (res < 0) return -1;
			len = sizeof(name);
			res = getsockname(sock, (struct sockaddr*) &name, &len);
			if (res < 0) {
				perror("getsockname");
				return -1;
			}
			port = ntohs(name.sin_port);
		}
		local_ports[sock] = port;
	}
	return port;
}


connection_info* get_conn_info(int sock, u_long ip, int port)
{
	int		loc_port;
	char		key[32];
	connection_info *c;

	loc_port = get_local_port(sock);
	sprintf(key, "%d:%lu:%d", loc_port, ip, port);
	c = hash_find(connections, key);
	if (c == NULL) {
		c = alloc(sizeof(connection_info));
		c->received = RM_MAX_ID;
		c->do_ack = FALSE;
		c->ack_index = -1;
		c->closing = FALSE;
		hash_insert(connections, key, c);
	}
	return c;
}


u_int new_seqno(int sock, u_long ip, int port)
{
	char	key[32];
	int	*p, loc_port;

	loc_port = get_local_port(sock);
	sprintf(key, "%d:%lu:%d", loc_port, ip, port);
	p = hash_find(seqno_table, key);
	if (p == NULL) {
		p = alloc(sizeof(int));
		*p = 0;
		hash_insert(seqno_table, key, p);
	}
	return (*p)++;
}


rm_msg* build_msg(int sock, char *data, int len, u_long ip, int port, int flags)
{
	u_int		seqno;
	rm_msg		*m;
	connection_info *c;

	seqno = new_seqno(sock, ip, port);
	c = get_conn_info(sock, ip, port);
	/* 
	 * Since this message will carry piggybacked ack,
	 * disable any pending asychronous ack to the 
	 * same host+port from this local port.
	 */
	c->do_ack = FALSE; 

	/* Create message and fill in fields */
	m = alloc(len + RM_HEADER_SIZE);
	m->id 		= seqno;
	m->type		= (flags & MSG_EOF) ? RM_EOF : RM_REG;
	m->len 		= len;
	m->rcvd 	= c->received; /* ack all msgs up to c->received */
	bcopy(data, m->data, len);

	return m;
}


void add_msg_to_tables(rm_msg* m, int s, u_long ip, int port, int flags)
{
	outstanding_msg	*o;

	o = alloc(sizeof(outstanding_msg));

	o->m 		= m;
	o->sock 	= s;
	o->flags 	= flags;
	o->ip 		= ip;
	o->port 	= port;
	o->next 	= NULL;
	o->prev 	= NULL;

	if (outs_h == NULL) {
		outs_h = o;
		outs_t = o;
	}
	else {
		o->prev = outs_t;
		outs_t->next = o;
		outs_t = o;
	}
}


/*
 * Receive an incoming message into two buffers without extra copying.
 */
int recv_vec(int sock, void* b1, int b1_len, void* b2, int b2_len, 
	     int flags, struct sockaddr_in* addr)
{
	struct msghdr	h;
	struct iovec	sg[2];
	int		val;

	sg[0].iov_base = b1;
	sg[0].iov_len = b1_len;
	sg[1].iov_base = b2;
	sg[1].iov_len = b2_len;

	h.msg_name = (caddr_t) addr;
	h.msg_namelen = (addr==NULL?0:sizeof(struct sockaddr_in));
	h.msg_iov = sg;
	h.msg_iovlen = 2;
	h.msg_control = NULL;
	h.msg_controllen = 0;
	h.msg_flags = 0;

	/*
	 * Retry this receive if it was interrupted by a signal
	 */
	while ((val = recvmsg(sock, &h, flags)) < 0 && errno == EINTR);
	return val;
}


int udp_send(int sock, void* data, int len, int flags, u_long ip, int port)
{
	int		res;
	struct timeval	t = {0, 1}; /* 1 us */
	struct sockaddr_in d;

	/* Fill in destination sockaddr struct */
	d.sin_family 		= AF_INET;
	d.sin_port 		= htons(port);
	d.sin_addr.s_addr 	= ip;

	/*
	 * The Freebsd messaging code easily fills up its available
	 * message buffers (~50 of them). If there are no available buffers
	 * we'll sleep for 1 clock tick and then try again.
	 */
	while ((res = sendto(sock, data, len, flags, (struct sockaddr*) &d, 
		sizeof(d))) < 0 && errno == ENOBUFS) {
		fprintf(stderr, ".");
		select(0, NULL, NULL, NULL, &t);
	}
	return res;
}


/* 
 * Is message 'id' within acked window? To be in the window, the message
 * must be between 'upto' and 'upto' - RM_WINDOW, even if that window wraps
 * around to higher ids.
 */
int in_wind(u_int id, u_int upto)
{
	u_int	lim;

	if (upto <= RM_WINDOW) {
		lim = RM_MAX_ID - (RM_WINDOW - upto);
		if (id <= upto || id >= lim) 
			return TRUE;
		else 
			return FALSE;
	}
	else {
		lim = upto - RM_WINDOW;
		if (id <= upto && id >= lim) 
			return TRUE;
		else 
			return FALSE;
	}
}


/*
 * process_acks() removes messages from retransmit list that have
 * been acknowledged. Signals should be off (rm_lock held) during this
 * call.
 */
void process_acks(u_long ip, int port, u_int upto)
{
	outstanding_msg *o, *n;
	int		c = 0;

	for (o = outs_h; o != NULL; o = n) {
		n = o->next;
		if (o->ip == ip && o->port == port && in_wind(o->m->id, upto)) {
			free(o->m); /* we're done with message */
			msgs_on_sock[o->sock]--; /* decr msgs/sock counter */
			c++;
			if (outs_h == o) {
				if (o->next != NULL)
					o->next->prev = NULL;
				outs_h = o->next;
				if (outs_h == NULL)
					outs_t = NULL;
			}
			else if (outs_t == o) {
				if (o->prev != NULL)
					o->prev->next = NULL;
				outs_t = o->prev;
			}
			else {
				o->prev->next = o->next;
				o->next->prev = o->prev;
			}
			free(o);
		}
	}
	fprintf(stderr, "\tprocess_acks: quenched %d msgs\n", c);
}

/*
 * flow_control() blocks until the number of outstanding messages on this
 * socket is below a compile time threshold. Signals must be enabled during
 * this call (rm_lock must not be held).
 */
void flow_control(int s)
{
	struct timeval	t = {0, 10000}; /* 10 ms */

	while (msgs_on_sock[s] >= RM_MAX_OUTSTANDING)
		select(0, NULL, NULL, NULL, &t);

	rm_lock();
	msgs_on_sock[s]++;
	rm_unlock();
}


/*
 * This function enables asynchronous acking for this connection.
 * A synchronous ack will go out for this connection at the next timer
 * signal, unless a regular message is sent on this connection first.
 * Signals must be disabled (rm_lock held) during this call.
 */
void async_ack_enable(int s, connection_info *c, u_long ip, int port)
{
	if (c->ack_index == -1) {
		if (curr_ack_index < RM_MAX_DESCRIPTORS) {
			c->ack_index = curr_ack_index;
			c->ip = ip;
			c->port = port;
			c->s = s;
			to_ack[curr_ack_index++] = c;
		}
	}
	c->do_ack = TRUE;
}


/*
 * This is our timer signal handler. All the asynchronous stuff is done here.
 * We send out singleton acks on connections where an ack has not recently
 * gone out (piggybacked on another message for example). We also check
 * for incoming acks at the heads of our socket buffers. Acks down in the
 * buffer behind a new message will not be caught until the new
 * message is consumed by the application. Finally, we retransmit any
 * recently sent messages that have not been acked.
 */
void handle_timer(int sig)
{
	int	i, len, res, s, port, done, connected;
	u_long	ip;
	rm_msg	ack, m;
	fd_set	socks;
	outstanding_msg *o;
	connection_info *c;
	struct timeval t = {0, 0};
	struct sockaddr_in f;
	int	addrlen = sizeof(f);
	static int sig_count = 0;
	int	acks_sent;
	int	retrans_sent;

	fprintf(stderr, "---------- in handler\n");

	/*
	 * Send explicit acks for messages whose acks have
	 * not yet been piggybacked.
	 */
	acks_sent = 0;
	for (i=0; i < curr_ack_index; i++) {
		c = to_ack[i];
		/* skip stuff that has already been acked */
		if (!c->do_ack) 
			continue;
			
		ack.id = 0;
		ack.type = RM_ACK;
		ack.len = 0;
		ack.rcvd = c->received;
		res = udp_send(c->s, &ack, RM_HEADER_SIZE, 0, c->ip, c->port);
		if (res < 0)
			perror("udp_send ack");
		acks_sent++;
	}
	/* Nullify ack list */
	for (i=0; i < curr_ack_index; i++) {
		to_ack[i]->ack_index = -1;
		to_ack[i] = NULL;
	}
	curr_ack_index = 0;

	if (acks_sent != 0)
		fprintf(stderr, "\tsent %d acks\n", acks_sent);

	/*
	 * Check the heads of all open socket queues for incoming acks
	 * or retransmissions. We will handle all acks and consume
	 * retransmissions. We will leave each socket either empty or
	 * with a new message at the head. We have to use select to make
	 * sure that we don't block while peeking the queues.
	 */
	done = FALSE;
	while (!done) {
		/*
		 * 'done' will stay TRUE and this loop will exit once
		 * we find no acks or retransmitted messages at the 
		 * head of any of our sockets.
		 */
		done = TRUE;
		FD_ZERO(&socks);
		for (i=0; i < num_desc; i++) {
			FD_SET(descriptors[i], &socks);
		}
		res = select(4096, &socks, NULL, NULL, &t);
		if (res <= 0) {
			break;
		}
		for (i=0; i < num_desc; i++) {
			if (!FD_ISSET(descriptors[i], &socks))
				continue;

			/* There is stuff to read on this socket. */
			s = descriptors[i];
			len = RM_HEADER_SIZE;
			res = recv_vec(s, &m, len, NULL, 0, MSG_PEEK, &f);
			if (res < 0) {
				perror("recv_vec in timer peek");
				continue;
			}

			ip = f.sin_addr.s_addr;
			port = ntohs(f.sin_port);
			c = get_conn_info(s, ip, port);
			if (m.type == RM_ACK) {
				/* handle this singleton ack */
				done = FALSE;
				fprintf(stderr, "\tgot singleton ack (%d)\n", m.rcvd);
				process_acks(ip, port, m.rcvd);
			}
			else if (!in_wind(m.id, c->received)) { /* ??? */
				/* A new message. Leave it in buffer */
				continue;
			}
			else {
				/* 
				 * This message is a retransmission.
				 * In case the ack for this message was lost,
				 * we'll turn on async acking for this msg. 
				 */
				done = FALSE;
				async_ack_enable(c->s, c, ip, port);
				fprintf(stderr, "\tconsuming %d (%d)\n", m.id, c->received+1);
			}

			/* Consume ack or retransmitted message */
			res = recv_vec(s, &m, len, NULL, 0, 0, &f);
			if (res < 0) {
				perror("recv_vec in timer");
				continue;
			}
		}
	}

	/*
	 * We wont retransmit messages every time this handler is invoked.
	 * Just every RM_SIGS_PER_RETRANS times.
	 */
	sig_count++;
	if (sig_count != RM_SIGS_PER_RETRANS) {
		fprintf(stderr, "---------- exiting handler early\n");
		return;
	}
	sig_count = 0;

	retrans_sent = 0;
	for (o = outs_h; o != NULL; o = o->next) {
		len = o->m->len + RM_HEADER_SIZE;
		res = udp_send(o->sock, o->m, len, o->flags, o->ip, o->port);
		if (res < 0) {
			perror("udp_send retrans");
		}
		retrans_sent++;
	}
	if (retrans_sent != 0)
		fprintf(stderr, "\tretransmitted %d messages\n",retrans_sent);

	fprintf(stderr, "---------- exiting handler\n");
}


/**
 ** Public Functions
 **/

/*
 * rm_init()
 *
 *	Set up rmessages.
 */
void rm_init(void)
{
	int	i;
	struct timeval duration = {0, RM_TIMER_PERIOD*1000};
	struct sigaction sa = {handle_timer, 0, SA_RESTART};

	seqno_table = alloc(sizeof(hash_table));
	hash_init(seqno_table, HASH_STRING_KEYS);
	connections = alloc(sizeof(hash_table));
	hash_init(connections, HASH_STRING_KEYS);
	addr_table = alloc(sizeof(hash_table));
	hash_init(addr_table, HASH_ONE_WORD_KEYS);

	outs_h = NULL;
	outs_t = NULL;

	for (i=0; i < RM_MAX_DESCRIPTORS; i++) {
		to_ack[i] = NULL;
		local_ports[i] = 0;
		msgs_on_sock[i] = 0;
	}
	curr_ack_index = 0;

	/* Set up timer to signal us every RM_TIMER_PERIOD ms */
	sigaction(RM_TIMER_SIG, &sa, NULL);
	if (fork() == 0) {
		int	parent, pid, res;

		parent = getppid();
		pid = getpid();
		for (;;) {
			res = select(0, NULL, NULL, NULL, &duration);
			if (res != 0) {
				fprintf(stderr, "timer child select failed\n");
				exit(1);
			}
			else {
				res = kill(parent, RM_TIMER_SIG);
				if (res < 0) {
					exit(0);
				}
			}
		}
		fprintf(stderr, "Timer child exiting?!?\n");
		exit(1);
	}
}


/*
 * rm_socket()
 *
 *	Return an open socket descriptor for use in later messaging calls,
 *	or -1 on failure.
 */
int rm_socket(void)
{
	int	d;

	if (num_desc == RM_MAX_DESCRIPTORS)
		return -1;

	rm_lock();
	d =  socket(AF_INET, SOCK_DGRAM, 0);
	if (d >= 0)
		descriptors[num_desc++] = d;
	rm_unlock();
	return d;
}


/*
 * rm_bind()
 *
 *	Bind a socket to a specific port. If the provided port number
 *	is zero, the system will choose an available port for the socket.
 *	Returns 0 on success or -1 on failure. 
 */
int rm_bind(int sock, int port)
{
	struct sockaddr_in addr;

	bzero(&addr, sizeof(addr));
	addr.sin_family 	= AF_INET;
	addr.sin_addr.s_addr 	= htonl(INADDR_ANY);
	addr.sin_port 		= htons(port);

	return bind(sock, (struct sockaddr*) &addr, sizeof(addr));
}


int rm_connect_ip(int sock, u_long ip, int port)
{
	int	res;
	struct sockaddr_in addr, *a;

	a = hash_find(addr_table, (char*)sock);
	if (a == NULL) {
		a = alloc(sizeof(struct sockaddr_in));
		hash_insert(addr_table, (char*)sock, a);
	}

	bzero(a, sizeof(*a));
	a->sin_family 		= AF_INET;
	a->sin_port 		= htons(port);
	a->sin_addr.s_addr 	= ip;

	return 0;
}


/*
 * rm_connect()
 *
 *	Connect a UDP socket to a named host. By connecting a connectionless
 *	socket in this manner, the user is really just specifying a 
 *	destination address so that later calls to send() and write()
 *	(functions that do not take a destination address parameter) know
 *	where to send their data. It returns 0 on success and -1 on failure.
 */
int rm_connect(int sock, char *host, int port)
{
	unsigned long ip;

	ip = get_address(host);
	if (!ip) 
		return -1;

	return rm_connect_ip(sock, ip, port);
}


/*
 * rm_fill_sockaddr()
 *
 *	This routine is provided as a convenience to the user. It takes
 *	a host address (a string consisting of a hostname or ip address 
 *	in dotted decimal format), and port and fills in a sockaddr_in
 *	struct appropriately. It returns -1 if the hostname is invalid
 *	and 0 otherwise.
 */
int rm_fill_sockaddr(struct sockaddr_in *a, char* host, int port)
{
	u_long	ip;

	ip = get_address(host);
	if (!ip)
		return -1;

	a->sin_family		= AF_INET;
	a->sin_port 		= htons(port);
	a->sin_addr.s_addr 	= ip;
	return 0;
}


/*
 * rm_sendto()
 *
 *	Sends a reliable datagram on a connected socket. Datagrams will be
 *	delivered on the receiving side in the same order in which they
 *	were sent, and packet losses will be invisible to the user. Returns
 *	0 on success and -1 on failure. 
 */
int rm_sendto(int sock, void *msg, int len, int flags, struct sockaddr_in *to)
{
	int	port, res;
	u_long	ip;
	rm_msg	*m;

	/* 
	 * If necessary, wait for consumer to catch up before 
	 * enqueueing new msg 
	 */
	flow_control(sock); 

	rm_lock();

	port = ntohs(to->sin_port);
	ip = to->sin_addr.s_addr;
	m = build_msg(sock, msg, len, ip, port, flags);
	add_msg_to_tables(m, sock, ip, port, flags);
	len += RM_HEADER_SIZE;

	rm_unlock();

	fprintf(stderr, "sent %d\n", m->id);
	return udp_send(sock, m, len, flags, ip, port);
}


/*
 * rm_recvfrom()
 *
 *	Receives a reliable datagram from a socket. These datagrams are
 *	delivered in the same order in which they were sent. Returns the
 *	number of bytes received on success and -1 on failure.
 */
int rm_recvfrom(int s, void *b, int len, int flags, struct sockaddr_in *from)
{
	int	res, port, addrlen;
	u_long	ip;
	rm_msg	m;
	connection_info *c;
	struct sockaddr_in addr;

	rm_lock();

	if (from == NULL) {
		from = &addr;
	}

	for (;;) {
		rm_unlock();
		res = recv_vec(s, &m, RM_HEADER_SIZE, b, len, flags, from);
		rm_lock();
		if (res < 0) {
			rm_unlock();
			return res;
		}

		port  	= ntohs(from->sin_port);
		ip	= from->sin_addr.s_addr;
		if (m.type == RM_ACK) {
			/*
			 * The message we just dequeued is 
			 * a singleton ack. If we only peeked the
			 * messasge, we'll consume it now.
			 */
			fprintf(stderr, "rf: got singleton ack\n");
			process_acks(ip, port, m.rcvd);
			if (flags & MSG_PEEK) {
				res = recv_vec(s, &m, RM_HEADER_SIZE, 
					       NULL, 0, 0, from);
				if (res < 0) {
					rm_unlock();
					return res;
				}
			}
		}
		else {
			/*
			 * We received a normal message. Handle ack and
			 * then see if this is the message we're waiting for
			 * from this host+port. If not, we'll drop it. 
			 * If the caller passed us the MSG_PEEK flag, don't
			 * increment thereceived seqno. That should be done 
			 * when the message is finally received.
			 */
			process_acks(ip, port, m.rcvd);
			c = get_conn_info(s, ip, port);
			if (m.id == c->received + 1) {
				/* 
				 * This is the message we're waiting for.
				 */
				if (flags & MSG_PEEK) {
					rm_unlock();
					return m.len;
				}

				fprintf(stderr, "rf: got %d\n", m.id);

				c->received++;

				/*
				 * Check if this is an EOF message. EOFs
				 * do not get automatically acked.
				 */
				if (m.type == RM_EOF) {
					c->closing = TRUE;
					c->do_ack = FALSE;
					rm_unlock();
					return m.len;
				}

				/*
				 * Set things up to ack this message if
				 * some other msg doesn't go back before
				 * the next timer signal.
				 */
				async_ack_enable(s, c, ip, port);

				/*
				 * We're done. Msg was copied to user 
				 * buffer by recv_vec.
				 */
				rm_unlock();
				return m.len;
			}
			else {
				/*
				 * This message is either out of order
				 * or one that we've already seen. We'll
				 * drop it. If we're 'peeking' for the user
				 * we'll need to deliberately consume
				 * this message. In case it's a retransmission
				 * and the original ack was lost, we'll enable
				 * another async ack.
				 */
				async_ack_enable(s, c, ip, port);
				fprintf(stderr, "rf: consuming %d (%d)\n", m.id, c->received+1);
				if (flags & MSG_PEEK) {
					res = recv_vec(s, &m, RM_HEADER_SIZE, 
						       NULL, 0, 0, from);
					if (res < 0) {
						rm_unlock();
						return res;
					}
				}
			}
		}
	}
}


/*
 * rm_recv()
 *
 *	Receives a reliable datagram from a socket. These datagrams are
 *	delivered in the same order in which they were sent. Returns the
 *	number of bytes received on success and -1 on failure.
 */
int rm_recv(int s, void *b, int len, int flags)
{
	return rm_recvfrom(s, b, len, flags, NULL);
}


/*
 * rm_send()
 *
 *	Sends a reliable datagram on a connected socket. Datagrams will be
 *	delivered on the receiving side in the same order in which they
 *	were sent, and packet losses will be invisible to the user. Returns
 *	0 on success and -1 on failure. 
 */
int rm_send(int sock, void *msg, int len, int flags)
{
	struct sockaddr_in *to;

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) {
		errno = ENOTCONN;
		return -1;
	}
	return rm_sendto(sock, msg, len, flags, to);
}


/*
 * rm_get_mtu()
 *
 *	Return the maximum number of bytes sendable in a single message 
 *	on a particular socket (the "maximum transfer unit" or MTU). The
 *	mtu is based on the sendbuffer size which the user can adjust
 * 	using the setsockopt() call.
 */
int rm_get_mtu(int s)
{
	int	len, res, bufsize, mtu;

	len = sizeof(bufsize);
	res = getsockopt(s, SOL_SOCKET, SO_SNDBUF, &bufsize, &len);
	if (res < 0) 
		return -1;
	else {
		mtu = bufsize - RM_HEADER_SIZE; 
		if (mtu > RM_MAX_MTU)
			return RM_MAX_MTU;
		else
			return mtu;
	}
}


/*
 * rm_bytes_ready()
 *
 *	This function returns the number of bytes that are available to be
 * 	returned on the given socket. It blocks until bytes are available.
 *	It consumes packets received out of order, as well as singleton
 *	ack messages. It returns the number of bytes available, or -1 on
 * 	failure.
 */
int rm_bytes_ready(int s)
{
	int	res, port;
	u_long	ip;
	rm_msg	m;
	connection_info *c;
	struct sockaddr_in from;

	rm_lock();

	for (;;) {
		rm_unlock();
		res = recv_vec(s, &m, RM_HEADER_SIZE, NULL, 0, MSG_PEEK, &from);
		rm_lock();
		if (res < 0) {
			rm_unlock();
			return res;
		}

		port  	= ntohs(from.sin_port);
		ip	= from.sin_addr.s_addr;
		if (m.type == RM_ACK) {
			/*
			 * The message we just dequeued is 
			 * a singleton ack. Consume it if it hasn't
			 * been consumed already.
			 */
			fprintf(stderr, "br: got singleton ack\n");
			process_acks(ip, port, m.rcvd);
			res = recv_vec(s, &m, RM_HEADER_SIZE, NULL, 0, 0,&from);
			if (res < 0) {
				rm_unlock();
				return res;
			}
		}
		else {
			c = get_conn_info(s, ip, port);
			if (m.id == c->received + 1) {
				/*
				 * This is the next message we're waiting
				 * for. Return it's length.
				 */
				rm_unlock();
				return m.len;
			}
			else {
				/*
				 * This message is either out of order
				 * or one that we've already seen. We'll
				 * drop it. The original ack for this 
				 * transmission was lost, we'll enable
				 * async ack for this connection.
				 */
				async_ack_enable(s, c, ip, port);

				fprintf(stderr, "br: consuming %d (%d)\n", m.id, c->received+1);
				res = recv_vec(s, &m, RM_HEADER_SIZE, 
					       NULL, 0, 0, &from);
				if (res < 0) {
					rm_unlock();
					return res;
				}
			}
		}
	}
}


/*
 * rm_drain()
 *
 *	Waits for all outgoing messages on 'sock' to be received and
 *	acknowledged by their recipients.
 */
void rm_drain(int sock)
{
	int		msgs, i;
	outstanding_msg *o;
	struct timeval	t = {0, 100000}; /* 100 ms */

	for (;;) {
		msgs = 0;
		rm_lock();
		for (o = outs_h; o != NULL; o = o->next) {
			if (o->sock == sock) msgs++;
		}
		rm_unlock();
		if (msgs == 0) 
			break;
		else {
			/*
			 * There is still stuff outstanding, so wait for a
			 * while then check again.
			 */
			select(0, NULL, NULL, NULL, &t);
		}
	}
}


/*
 * rm_close()
 *
 * 	Close a socket. This function cleans up internal data structures
 *	and attempts to ensure that all the messages sent on this socket
 *	have been received.
 */
int rm_close(int sock)
{
	int		msgs, i;
	outstanding_msg	*o;
	struct timeval	t = {0, 100000};

	rm_lock();

	/*
	 * We don't have enough information here to remove the entries
	 * in the "connections" table and the "seqno" table, each of which
	 * are indexed by the full (sock, ip, port) of the peer. We'll
	 * assume here that if a new connection if opened to the same peer,
	 * it will at least have a new port number (possibly also a new 
	 * socket number), so that new clean entries in those two tables
	 * will be automatically created. 
	 */

	/*
	 * Remove address info from connection address table.
	 */
	hash_delete_entry(addr_table, (char*)sock);

	/*
	 * Reset the port & msgs count associated with this socket.
	 */
	local_ports[sock] = 0;
	msgs_on_sock[sock] = 0;

	/*
	 * Remove this socket from our list of open fds.
	 */
	for (i=0; i < num_desc; i++)
		if (descriptors[i] == sock) 
			break;
	if (i == num_desc)
		fprintf(stderr, "warning: socket fd not found in open fds.\n");
	else {
		for (i++; i < num_desc; i++)
			descriptors[i] = descriptors[i+1];
		num_desc--;
	}

	rm_unlock();

	return close(sock);
}


/*
 * rm_conn_closing()
 *
 *	Returns true if sock's connection has already received an EOF
 *	message, and false otherwise. Returns FALSE if called on a
 *	non-connected socket.
 */
int rm_conn_closing(int sock)
{
	u_long		ip;
	int		port;
	connection_info	*c;
	struct sockaddr_in *to;

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) 
		return FALSE;
	
	ip = to->sin_addr.s_addr;
	port = ntohs(to->sin_port);
	c = get_conn_info(sock, ip, port);
	if (c == NULL)
		return FALSE;

	return c->closing;
}


/*
 * rm_ack()
 *
 *	This function sends a current singleton ack on the provided
 *	connected socket.
 */
int rm_ack(int sock)
{
	u_long		ip;
	int		port;
	rm_msg		m;
	connection_info	*c;
	struct sockaddr_in *to;

	to = hash_find(addr_table, (char*)sock);
	if (to == NULL) {
		errno = ENOTCONN;
		return -1;
	}
	
	ip = to->sin_addr.s_addr;
	port = ntohs(to->sin_port);
	c = get_conn_info(sock, ip, port);

	/*
	 * Manually build a singleton ack.
	 */
	m.id 	= 0;
	m.type 	= RM_ACK;
	m.len 	= 0;
	m.rcvd 	= c->received;

	/*
	 * Send it unreliably
	 */
	return udp_send(sock, &m, RM_HEADER_SIZE, 0, ip, port);
}


