/*
 * Rmessage library. By Dave Lowell.
 */

#ifndef RMESSAGE_H
#define RMESSAGE_H

/**
 ** Includes
 **/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <limits.h>
#include "hash.h"

/**
 ** Defines
 **/

#define TRUE 1
#define FALSE 0

#define RM_MAX_DESCRIPTORS 	1024
#define RM_MAX_ID 		UINT_MAX
#define RM_WINDOW 		(RM_MAX_ID/2)
#define RM_REG 			1
#define RM_ACK 			2
#define RM_NACK 		3
#define RM_EOF 			4
#define RM_TIMER_SIG		SIGSYS
#define RM_MAX_MTU 		40000

/* retransmission specifiers */
#define RM_RETRANS_ALL		1
#define RM_RETRANS_HOST		2
#define RM_RETRANS_MSG		3

/* Tunable constants */

#define RM_TIMER_PERIOD		125 		/* period in ms */
#define RM_SIGS_PER_RETRANS	16		/* retrans every N timer ints */

/* Interhost traffic */
#define RM_OUT_HIGH 		100		/* Outstanding msg high water */
#define RM_OUT_LOW 		30		/* Outstanding msg low water */
#define RM_OUT_BYTES_HIGH 	(128*1024)	/* Out. bytes/sock high water */
#define RM_OUT_BYTES_LOW 	(32*1024)	/* Out. bytes/sock low water */
#define RM_ACK_EVERY 		4

#ifdef UNDEF
/* Intrahost traffic */ /* set for 240k buffer && 8k msgs */
#define RM_OUT_HIGH 		16		/* Outstanding msg high water */
#define RM_OUT_LOW 		6		/* Outstanding msg low water */
#define RM_MAX_OUT_BYTES 	(128*1024)	/* max outstanding bytes/sock */
#define RM_ACK_EVERY 		(RM_OUT_HIGH/4) 
#endif

/*
#define RM_ACK_EVERY 		1 
*/


/**
 ** Types
 **/

/* rm message header */
typedef struct {
	u_int	id;		/* mesg id. Used for ordering */
	short	type;		/* type of packet: ack or regular */
	short	len;		/* size of user data */
	u_int	rcvd;		/* ack up to mesg # 'rcvd' */
	short	retrans;
	char	data[0];	/* variable length user data */
} rm_msg;
#define RM_HEADER_SIZE (sizeof(rm_msg))

/* incoming message we've received and sender address */
typedef struct mbuf_s {
	rm_msg		*m;
	struct sockaddr_in from;
	struct mbuf_s 	*next;
} mbuf;

/* per socket in-order list of messages to be delivered to app */
typedef struct sockbuf_s {
	mbuf	*head;		/* head pointer for list of incoming msgs */
	mbuf	*tail;		/* tail pointer for list of incoming msgs */
} sockbuf;

/* data per socket-host-port */
typedef struct {
	u_int	received;	/* largest msg id received in sequence */
	int	mcount;		/* count of messages received since last ack */
	int	do_ack;		/* send ack if true */
	int	ack_index;	/* our slot in to_ack list */
	u_int	nack_seq;	/* nack suppression info */
	int	s;		/* sock for sending singleton acks */
	u_long	ip;		/* dest ip for sending singleton acks */
	int	port;		/* dest port for sending singleton acks */
	int	closing;	/* flag to let us know if we've received EOF */
	hash_table *oo_table;	/* table of out-of-order messages */
} connection_info;

/* message that may need to be retransmitted */
typedef struct out_msg_s {
	rm_msg	*m;		/* msg that was sent */
	int	sock;		/* socket on which it was sent */
	int	flags;		/* flags under which is was sent */
	u_long	ip;		/* host to which it was sent */
	int	port;		/* port on above host to which it was sent */
	int	ticks;		/* clock ticks counter */
	struct out_msg_s *next;	/* next msg on list */
	struct out_msg_s *prev;	/* previous msg on list */
} outstanding_msg;

typedef struct {
	int	id;
	char	retrans;
} tracent;


/**
 ** Public Functions
 **/

extern void rm_init(void);
extern int rm_socket(void);
extern int rm_bind(int, int);
extern int rm_connect(int, char*, int);
extern int rm_connect_ip(int, u_long, int);
extern int rm_sendto(int, void*, int, int, struct sockaddr_in*);
extern int rm_recvfrom(int, void*, int, int, struct sockaddr_in*);
extern rm_msg* rm_recvp(int, int, struct sockaddr_in*);
extern int rm_recv(int, void*, int, int);
extern int rm_send(int, void*, int, int);
extern int rm_fill_sockaddr(struct sockaddr_in*, char*, int);
extern int rm_get_mtu(int);
extern int rm_bytes_ready(int);
extern void rm_drain(int);
extern int rm_close(int);
extern int rm_conn_closing(int);
extern int rm_ack(int);

#endif

